/* 
 * ------------------------------------------------------------------
 * QWKToSpool.c - Convert QWK file to news spool
 * Created by Robert Heller on Sun Nov 12 13:11:02 1995
 * ------------------------------------------------------------------
 * Modification History:
 * ------------------------------------------------------------------
 * Contents:
 * ------------------------------------------------------------------
 *  
 * 
 * Copyright (c) 1995 by Robert heller
 *        All Rights Reserved
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <regex.h>
#include <errno.h>
#include <sys/stat.h>
#include "qwk.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

static char userName[50];

static char *fgetscrlf(char *buffer,int bsize,FILE *fp)
{
	char *result = buffer;
	while (bsize > 2)
	{
		int c = getc(fp);
		if (c < 0)
		{
			if (result == buffer) return(NULL);
			*buffer = '\0';
			return (result);
		} else if (c == '\r')
		{
			c = getc(fp);
			if (c == '\n')
			{
				*buffer++ = '\n';
				bsize--;
				if (bsize > 0) *buffer = '\0';
				return (result);
			} else if (c != '\n')
			{
				*buffer++ = '\r';
				bsize--;
			} else if (c < 0)
			{
				*buffer = '\0';
				return (result);
			}
			*buffer++ = c;
			bsize--;
		} else
		{
			*buffer++ = c;
			bsize--;
		}
	}
	return (result);
}

void mkConfDir(const char *spoolDirectory,const char *group,char *dirbuffer)
{
	char *p,*g,*d;

	strcpy(dirbuffer,spoolDirectory);
	if (mkdir(dirbuffer,S_IREAD|S_IWRITE|S_IEXEC) < 0)
	{
		int err = errno;
		if (err != EEXIST)
		{
			perror("mkdir");
			exit(err);
		}
	}
	g = (char*)group;
	d = dirbuffer + strlen(dirbuffer);
	while (*g != '\0')
	{
		*d++ = '/';
		while (*g != '.' && *g != '\0') *d++ = *g++;
		*d = '\0';
		if (mkdir(dirbuffer,S_IREAD|S_IWRITE|S_IEXEC) < 0)
		{
			int err = errno;
			if (err != EEXIST)
			{
				perror("mkdir");
				exit(err);
			}
		}
		if (*g == '.') g++;
	}
}

int
readCnum(const byte * ptr)
{
	unsigned char   p, q;
	unsigned int    j;

	p = *ptr;
	ptr++;
	q = *ptr;
	q = q & 0x1f;
	j = (unsigned int) q;
	j <<= 8;
	j += (unsigned int) p;

	return (j);
}

Conferences *confs = NULL;

int lastconf = -1;

void ReadControl(const char *controlFile, const char *spoolDirectory)
{
	FILE *ctrlIn;
	static char lineBuffer[256];
	char *lp, *p;
	int iconf, i;
	int mm,dd,year,hh,mn,sec;

	if ((ctrlIn = fopen(controlFile,"rb")) == NULL)
	{
		int err = errno;
		perror("fopen: controlFile");
		exit(err);
	}
	lp = fgetscrlf(lineBuffer,256,ctrlIn);	/* board name */
	if (lp == NULL)
	{
		int err = errno;
		perror("fgetscrlf: board name");
		exit(err);
	}
	for (i = 0;i < 5;i++)
	{
		lp = fgetscrlf(lineBuffer,256,ctrlIn);
		if (lp == NULL)
		{
			int err = errno;
			perror("fgetscrlf: skip lines");
			exit(err);
		}
	}
	sscanf(lineBuffer,"%d-%d-%d,%d:%d:%d", &mm, &dd, &year, &hh, &mn, &sec);
	lp = fgetscrlf(lineBuffer,256,ctrlIn);  /* user name */
	if (lp == NULL)
	{
		int err = errno;
		perror("fgetscrlf: user name");
		exit(err);
	}
	memset(userName,'\0',50);
	strncpy(userName,lineBuffer,50);
	userName[49] = '\0';
	p = strchr(userName,'\n');
	if (p != NULL) *p-- = '\0';
	else p = userName+strlen(userName)-1;
	while (*p == ' ') *p-- = '\0';
	for (i = 0;i < 3;i++)
	{
		lp = fgetscrlf(lineBuffer,256,ctrlIn);
		if (lp == NULL)
		{
			int err = errno;
			perror("fgetscrlf: skip lines 2");
			exit(err);
		}
	}
	lp = fgetscrlf(lineBuffer,256,ctrlIn);  /* last conf */
	if (lp == NULL)
	{
		int err = errno;
		perror("fgetscrlf: last conf");
		exit(err);
	}
	lastconf = atoi(lp);
	confs = (Conferences*) malloc(sizeof(Conferences) * (lastconf+2));
	if (confs == NULL)
	{
		int err = errno;
		perror("malloc: Conferences");
		exit(err);
	}
	for (i = 0; i <= lastconf;i++)
	{
		lp = fgetscrlf(lineBuffer,256,ctrlIn);
		if (lp == NULL)
		{
			int err = errno;
			perror("fgetscrlf: conf number");
			exit(err);
		}
		confs[i].number = atoi(lp);
		lp = fgetscrlf(lineBuffer,256,ctrlIn);
		if (lp == NULL)
		{
			int err = errno;
			perror("fgetscrlf: conf name");
			exit(err);
		}
		p = strchr(lp,'\n');
		if (p != NULL) *p = '\0';
		strncpy(confs[i].name,lp,64);
		confs[i].name[63] = '\0';
		confs[i].activeEntry = NULL;
	}
	fclose(ctrlIn);
	confs[lastconf+1].number = REPLYCNUM;
	strcpy(confs[lastconf+1].name,REPLY);
	lastconf += 2;
	for (i = 0; i < lastconf; i++)
	{
		mkConfDir(spoolDirectory,confs[i].name,confs[i].directory);
	}
}

Active *activeTable = NULL;
int numActive = 0,maxActive = 0;

static int blankLine(const char *lp)
{
	if (lp == NULL) return 1;
	while (*lp <= ' ' && *lp != '\0' && *lp != '\n') lp++;
	if (*lp == '\n' || *lp == '\0') return 1;
	return 0;
}

void ReadActive(const char *activeFile)
{
	FILE *activeIn;
	static char lineBuffer[256];
	char *lp, *p;
	char *n, *f;
	int ilast, ifirst;
	int i;
	numActive = 0;
	if ((activeIn = fopen(activeFile,"r")) == NULL) return;
	maxActive = lastconf;
	if (maxActive < 0) maxActive = 0;
	while (fgets(lineBuffer,256,activeIn) != NULL &&
	       !blankLine(lineBuffer)) maxActive++;
	rewind(activeIn);
	activeTable = (Active *) malloc(sizeof(Active)*maxActive);
	if (activeTable == NULL)
	{
		int err = errno;
		perror("malloc: Active");
		exit(err);
	}
	while ((lp = fgets(lineBuffer,256,activeIn)) != NULL)
	{
		if (blankLine(lp)) continue;
		while (*lp <= ' ' && *lp != '\0' && *lp != '\n') lp++;
		if (*lp == '\n' || *lp == '\0') continue;
		n = lp;
		p = strchr(lp,' ');
		if (p == NULL) continue;
		*p++ = '\0';
		ilast = atoi(p);
		p = strchr(p,' ');
		if (p == NULL) continue;
		*p++ = '\0';
		ifirst = atoi(p);
		f = strchr(p,' ');
		if (f == NULL) continue;
		f++;
		p = strchr(p,'\n');
		if (p == NULL) continue;
		*p = '\0';
		if (*f == '=')
		{
			f++;
			for (i = 0; i < numActive; i++)
			{
				if (strcmp(activeTable[i].name,f) == 0)
				{
					activeTable[i].confNumber = atoi(n);
					break;
				}
			}
			if (i == numActive)
			{
				numActive++;
				activeTable[i].confNumber = atoi(n);
				strcpy(activeTable[i].name,f);
				activeTable[i].last = 0;
				activeTable[i].first = 1;
				activeTable[i].newsrcFlag = FALSE;
			}
		} else
		{
			for (i = 0; i < numActive; i++)
			{
				if (strcmp(activeTable[i].name,n) == 0)
				{
					activeTable[i].last = ilast;
					activeTable[i].first = ifirst;
					break;
				}
			}
			if (i == numActive)
			{
				numActive++;
				strcpy(activeTable[i].name,f);
				activeTable[i].last = ilast;
				activeTable[i].first = ifirst;
				activeTable[i].newsrcFlag = FALSE;
			}
		}
	}
	fclose(activeIn);
}

void BuildActiveAndConfs(const char *controlFile, const char *spoolDirectory,const char *activeFile)
{
	int ic,ia;

	ReadControl(controlFile,spoolDirectory);
	ReadActive(activeFile);
	if (maxActive == 0)
	{
		maxActive = lastconf;
		activeTable = (Active *) malloc(sizeof(Active)*maxActive);
 		if (activeTable == NULL)
		{
			int err = errno;
			perror("malloc: Active");
			exit(err);
		}
	}
	for (ic = 0; ic < lastconf; ic++)
	{
		for (ia = 0;ia < numActive; ia++)
		{
			if (strcmp(activeTable[ia].name,confs[ic].name) == 0)
			{
				confs[ic].activeEntry = &activeTable[ia];
				activeTable[ia].confNumber = confs[ic].number;
				break;
			}
		}
		if (ia == numActive)
		{
			numActive++;
			strcpy(activeTable[ia].name,confs[ic].name);
			activeTable[ia].last = 0;
			activeTable[ia].first = 1;
			confs[ic].activeEntry = &activeTable[ia];
			activeTable[ia].confNumber = confs[ic].number;
			activeTable[ia].newsrcFlag = FALSE;
		}
	}
}

long SizeOfFile(const char *messageDatFile)
{
	FILE *tmp;
	long EOFPos;

	if ((tmp = fopen(messageDatFile,"rb")) == NULL) return 0L;
	if (fseek(tmp,0L,SEEK_END) < 0) {
		fclose(tmp);
		return 0L;
	}
	EOFPos = ftell(tmp);
	fclose(tmp);
	return EOFPos;
}

typedef struct {
	enum {FROM,SUBJECT} field;
	char *patternBuffer;
	regex_t pattern;
} KillPattern;

#define KILLPATTERNALLOCSIZE 100

int ReadKillFile(const char *killfile,KillPattern **patterns)
{
	static regex_t commentPattern, linePattern;
	static regmatch_t matched[3];
	int patternCount = 0, patternAlloc = 0;
	char *pattern,*q,*p;
	int status;
	static char errbuffer[512];
	FILE *kfp;
	static char oline[2048],line[2048];

	status = regcomp(&commentPattern,"#.*$",REG_NEWLINE|REG_EXTENDED);
	if (status != 0)
	{
		regerror(status,&commentPattern,errbuffer,512);
		fprintf(stderr,"ReadKillFile: regcomp(commentPattern): %s\n",errbuffer);
		exit(status);
	}
	status = regcomp(&linePattern,"^(from|subject):[[:space:]]*(.*)$",REG_NEWLINE|REG_EXTENDED|REG_ICASE);
	if (status != 0)
	{
		regerror(status,&linePattern,errbuffer,512);
		fprintf(stderr,"ReadKillFile: regcomp(linePattern): %s\n",errbuffer);
		exit(status);
	}
	kfp = fopen(killfile,"r");
	if (kfp == NULL) return 0;
	while (fgets(oline,sizeof(oline),kfp) != NULL) {
		if ((status = regexec(&commentPattern,oline,1,matched,0)) == 0) {
			strncpy(line,oline,matched[0].rm_eo-1);
			line[matched[0].rm_eo-1] = '\n';
			line[matched[0].rm_eo] = '\0';
		} else {
			strcpy(line,oline);
		}
		for (q = line; *q != '\0' && *q <= ' '; q++) ;
		if (q > line) {
			p = line;
			while (*q != '\0') {*p++ = *q++;}
		}
		q = strrchr(line,'\n');
		if (q != NULL) {
			for (q--;q >= line && *q <= ' '; q--) ;
			*++q = '\n';
			*++q = '\0';
		}
		if (line[0] == '\n' || line[0] == '\0') continue;
		if ((status = regexec(&linePattern,line,3,matched,0)) == 0) {
			line[matched[1].rm_eo] = '\0';
			p = &(line[matched[2].rm_so]);
			q = strrchr(p,'\n');
			if (q != NULL) *q = '\0';
			if (++patternCount > patternAlloc) {
				if (patternAlloc == 0) {
					*patterns = malloc(sizeof(KillPattern)*KILLPATTERNALLOCSIZE);
					if (*patterns == NULL) {
						int err = errno;
						perror("ReadKillFile: malloc(KillPattern)");
						exit(err);
					}
					patternAlloc = KILLPATTERNALLOCSIZE;
				} else {
					*patterns = realloc(*patterns,sizeof(KillPattern)*(KILLPATTERNALLOCSIZE+patternAlloc));
					if (*patterns == NULL) {
						int err = errno;
						perror("ReadKillFile: malloc(KillPattern)");
						exit(err);
					}
					patternAlloc += KILLPATTERNALLOCSIZE;
				}
			}						
			if (line[0] == 'f' || line[0] == 'F') {
				(*patterns)[patternCount-1].field = FROM;
			} else {
				(*patterns)[patternCount-1].field = SUBJECT;
			}
			(*patterns)[patternCount-1].patternBuffer = malloc(sizeof(char)*(strlen(p)+1));
			if ((*patterns)[patternCount-1].patternBuffer == NULL) {
				int err = errno;
				perror("ReadKillFile: malloc(patternBuffer)");
				exit(err);
			}
			strcpy((*patterns)[patternCount-1].patternBuffer,p);
			status = regcomp(&((*patterns)[patternCount-1].pattern),
					 (*patterns)[patternCount-1].patternBuffer,
					 REG_EXTENDED);
			if (status != 0) {
				fprintf(stderr,"ReadKillFile: %s: Bad regular expression at  %s",
					killfile,oline);
			}
		} else {
			fprintf(stderr,"ReadKillFile: %s: Syntax error at %s",
				killfile,oline);
		}
	}
	fclose(kfp);
	regfree(&commentPattern);
	regfree(&linePattern);
	return patternCount;
}

int PassKillFile(const char *messageFile,int patternCount,KillPattern *patterns)
{
	FILE *mfp;
	int kpindex,status;
	char headerline[2048];
	
	mfp = fopen(messageFile,"r");
	if (mfp == NULL) return FALSE;
	while (fgets(headerline,sizeof(headerline),mfp) != NULL) {
		if (headerline[0] == '\n') break;
		if (strncasecmp("from: ",headerline,6) == 0) {
			for (kpindex = 0;kpindex < patternCount;kpindex++) {
				if (patterns[kpindex].field != FROM) continue;
				status = regexec(&(patterns[kpindex].pattern),&headerline[6],0,NULL,0);
				if (status == 0) {
					fclose(mfp);
#ifdef DEBUG
					fprintf(stderr,"*** PassKillFile: failed: %s",headerline);
#endif
					return FALSE;
				}
			}
		} else if (strncasecmp("subject: ",headerline,9) == 0) {
			for (kpindex = 0;kpindex < patternCount;kpindex++) {
				if (patterns[kpindex].field != SUBJECT) continue;
				status = regexec(&(patterns[kpindex].pattern),&headerline[9],0,NULL,0);
				if (status == 0) {
					fclose(mfp);
#ifdef DEBUG
					fprintf(stderr,"*** PassKillFile: failed: %s",headerline);
#endif
					return FALSE;
				}
			}
		}
	}
	fclose(mfp);
#ifdef DEBUG
	fprintf(stderr,"*** PassKillFile: passed: %s\n",messageFile);
#endif
	return TRUE;					
}

void ProcessMESSAGESDAT(const char *messageDatFile,const char *killfile)
{
	FILE *messageDatIn, *spoolFileOut;
	KillPattern *killPatterns = NULL;
	int numberOfKillPatterns = 0;
	struct MsgHeaderType MessageHeader;
	static char buffer[128], filename[256];
	char *p,*colon;
	int count, oldcount, kcount;
	int conf = -1, cnum, iconf,i;
	int iconference,numBlocks, headerP;
	Active *current;
	int messageNumber;
	double SizeOfmessageDatFile, PosOfmessageDatFile;
	double Percent;

	if (strlen(killfile) > 0) {
		numberOfKillPatterns = ReadKillFile(killfile,&killPatterns);
	}	

	SizeOfmessageDatFile = (double) SizeOfFile(messageDatFile);

	if ((messageDatIn = fopen(messageDatFile,"rb")) == NULL)
	{
		int err = errno;
		perror("fopen: messageDatFile");
		exit(err);
	}
	if (fread(buffer,sizeof(buffer),1,messageDatIn) != 1)
	{
		int err = errno;
		perror("fread: header0");
		exit(err);
	}
	while (1)
	{
		if (fread(&MessageHeader.Status,sizeof(struct MsgHeaderType),1,messageDatIn) != 1)
		{
			break;
		}
		if (MessageHeader.Status == 0) break;
		cnum = readCnum((byte*)&(MessageHeader.BinConfN));
		if (cnum != conf)
		{
			if (conf >= 0)
			{
				printf("Conference %d \t[ %-63s ] %3ld New Message%s\n",
					confs[iconf].number,confs[iconf].name,
					count - oldcount,(count - oldcount != 1)?"s":"");
				if (kcount > 0) {
					printf("\t(%d articles killed)\n",kcount);
				}
			}
			count = 0;
			kcount = 0;
			oldcount = 0;
			conf = cnum;
			for (iconf = 0;iconf < lastconf;iconf++)
			{
				if (conf == confs[iconf].number) break;
			}
			if (iconf == lastconf)
			{
				MessageHeader.BinConfN[1] = 0;
				cnum = readCnum((byte*)&(MessageHeader.BinConfN));
				for (iconf = 0;iconf < lastconf;iconf++)
				{
					if (conf == confs[iconf].number) break;
				}
				if (iconf == lastconf)
				{
					iconf = 0;
					conf = confs[iconf].number;
				}
			}
			current = confs[iconf].activeEntry;
			oldcount = current->last;
			count = oldcount;
		}
		messageNumber = ++current->last;
		count++;
		sprintf(filename,"%s/%d",confs[iconf].directory,messageNumber);
		spoolFileOut = fopen(filename,"w");
		if (spoolFileOut == NULL)
		{
			int err = errno;
			perror("fopen: spoolFileOut");
			exit(err);
		}
		strncpy(buffer,MessageHeader.ForWhom,25);
		buffer[25] = '\0';
		for (p = &buffer[24]; *p == ' ' && p > buffer; p--) *p = '\0';
		/*fprintf(spoolFileOut,"X-QWK-To: %s\n",buffer);*/
		strncpy(buffer,MessageHeader.Author,25);
		buffer[25] = '\0';
		for (p = &buffer[24]; *p == ' ' && p > buffer; p--) *p = '\0';
		/*fprintf(spoolFileOut,"X-QWK-From: %s\n",buffer);*/
		strncpy(buffer,MessageHeader.Subject,25);
		buffer[25] = '\0';
		for (p = &buffer[24]; *p == ' ' && p > buffer; p--) *p = '\0';
		/*fprintf(spoolFileOut,"X-QWK-Subject: %s\n",buffer);*/
		strncpy(buffer,MessageHeader.MsgDate,8);
		buffer[8] = ' ';
		strncpy(buffer+9,MessageHeader.MsgTime,5);
		buffer[8+1+5] = '\0';
		/*fprintf(spoolFileOut,"X-QWK-Date: %s\n",buffer);*/
		headerP = FALSE;
		colon = NULL;
		numBlocks = atoi(MessageHeader.SizeMsg) - 1;
		for (i = 0; i < numBlocks; i++)
		{
			if (fread(buffer,sizeof(buffer),1,messageDatIn) != 1)
			{
				int err = errno;
				perror("fread: message");
				exit(err);
			}
			for (p = buffer;p != NULL;)
			{
				p = strchr(p,(char) 227);
				if (p != NULL) *p = '\n';
			}
			if (!headerP)
			{
				if (colon == NULL) colon = strchr(buffer,':');
				else colon -= sizeof(buffer);
				p = strchr(buffer,'\n');
				if (colon != NULL && p != NULL && colon > p)
				{
					fputc('\n',spoolFileOut);
					headerP = TRUE;
				} else if (p != NULL && colon == NULL)
				{
					fputc('\n',spoolFileOut);
					headerP = TRUE;
				}
			}
			if (fwrite(buffer,sizeof(buffer),1,spoolFileOut) != 1)
			{
				int err = errno;
				perror("fwrite: message");
				exit(err);
			}
		}
		fclose(spoolFileOut);
		if (!PassKillFile(filename,numberOfKillPatterns,killPatterns)) {
		  current->last--;
		  count--;
		  kcount++;
		}
		PosOfmessageDatFile = (double) ftell(messageDatIn);
		Percent = PosOfmessageDatFile / SizeOfmessageDatFile;
		printf("###%f\n",Percent);
	}
	if (Percent < 1.0) printf("###1.0\n");
	fclose(messageDatIn);
	if (conf >= 0)
	{
		printf("Conference %d \t[ %-63s ] %3ld New Message%s\n",
			confs[iconf].number,confs[iconf].name,
			count - oldcount,(count - oldcount != 1)?"s":"");
	}
}

void WriteActive(const char *activeFile,const char *newsrc)
{
	static char newFile[256],backupFile[256];
	FILE *in,*out;
	static char line[256];
	char *p;
	int i;

	sprintf(newFile,"%s.new",activeFile);
	sprintf(backupFile,"%s~",activeFile);
	out = fopen(newFile,"w");
	if (out == NULL)
	{
		int err = errno;
		perror("fopen: newActive");
		exit(err);
	}
	for (i = 0; i < numActive; i++)
	{
		fprintf(out,"%d %0.10d %05d =%s\n%s %0.10d %05d %c\n",
			    activeTable[i].confNumber,activeTable[i].last,
			    activeTable[i].first,activeTable[i].name,
			    activeTable[i].name,activeTable[i].last,
			    activeTable[i].first,
			    (strcmp(activeTable[i].name,REPLY) == 0)?'n':'y');
	}
	fclose(out);
	if (access(activeFile,F_OK) == 0)
	{
		rename(activeFile,backupFile);
	}
	if (rename(newFile,activeFile) != 0)
	{
		int err = errno;
		perror("rename: newActive=>activeFile");
		exit(err);
	}
	if (access(newsrc,F_OK) < 0)
	{
		out = fopen(newsrc,"w");
		if (out == NULL)
		{
			int err = errno;
			perror("fopen: newsrc");
			exit(err);
		}
		for (i = 0; i < numActive; i++)
		{
			fprintf(out,"%s:\n",activeTable[i].name);
		}
		fclose(out);
	} else
	{
		sprintf(newFile,"%s.new",newsrc);
		sprintf(backupFile,"%s~",newsrc);
		in = fopen(newsrc,"r");
		if (in == NULL)
		{
			int err = errno;
			perror("fopen: oldnewsrc");
			exit(err);
		}
		out = fopen(newFile,"w");
		if (out == NULL)
		{
			int err = errno;
			perror("fopen: newnewsrc");
			exit(err);
		}
		while (fgets(line,256,in) != NULL)
		{
			fputs(line,out);
			p = strchr(line,':');
			if (p == NULL) p = p = strchr(line,'!');
			if (p == NULL) continue;
			*p = '\0';
			for (i = 0; i < numActive; i++)
			{
				if (strcmp(activeTable[i].name,line) == 0)
				{
					activeTable[i].newsrcFlag = TRUE;
					break;
				}
			}
		}
		fclose(in);
		for (i = 0; i < numActive; i++)
		{
			if (!activeTable[i].newsrcFlag)
			{
				fprintf(out,"%s:\n",activeTable[i].name);
			}
		}
		fclose(out);
		rename(newsrc,backupFile);
		if (rename(newFile,newsrc) < 0)
		{
			int err = errno;
			perror("rename: newnewsrc=>newsrc");
			exit(err);
		}
	}
}

int main(int argc,char *argv[])
{
	char *control, *messages, *active, *spool, *newsrc, *killfile;

	if (argc < 6 || argc > 7)
	{
		fprintf(stderr,"usage: QWKToSpool control messages active spool newsrc [killfile]\n");
		exit(1);
	}
	control = argv[1];
	messages = argv[2];
	active = argv[3];
	spool = argv[4];
	newsrc = argv[5];
	if (argc == 7) {
		killfile = argv[6];
	} else {
		killfile = "";
	}
	BuildActiveAndConfs(control,spool,active);
	ProcessMESSAGESDAT(messages,killfile);
	WriteActive(active,newsrc);
	printf("\nUsername:%s\n",userName);
	exit(0);
}
